from pathlib import Path
from typing import Any, Iterator, Self

import pytest
from _pytest.nodes import Node

from pytest_stochastics.runner_data import Policy


class StochasticFunctionCollector(pytest.Collector):
    """Collector for stochastic tests."""

    def __init__(
        self,
        name: str,
        policy: Policy,
        obj: object,
        results: list[bool] = list(),
        parent: Node | None = None,
        config: pytest.Config | None = None,
        session: pytest.Session | None = None,
        path: Path | None = None,
        nodeid: str | None = None,
    ) -> None:
        self.obj = obj
        self.policy = policy
        self.results = results
        super().__init__(name, parent, config, session, None, path, nodeid)

    @classmethod
    def from_parent(
        cls,
        parent: Node,
        **kwargs: Any,
    ) -> Self:
        """Cooperative constructor which pytest likes, use this instead of `__init__`."""

        return super().from_parent(parent, **kwargs)  # type: ignore

    def collect(self) -> Iterator[pytest.Item]:
        """Collect `out_of` copies of the function base of the strategy."""

        for i in range(self.policy.out_of):
            func_name = f"{i+1:>02d} of {self.policy.out_of:>02d}" if self.policy.out_of > 1 else self.name
            yield StochasticFunction.from_parent(self, name=func_name, callobj=self.obj)  # type: ignore


class StochasticFunction(pytest.Function):
    """Marker that we wrap a `pytest.Function`."""

    pass



import logging
from logging import Logger
import rich

import pytest

from pytest_stochastics.runner import RunnerStochastics
from pytest_stochastics.runner_data import PlanId, RunnerStochasticsConfig

PYTEST_STOCHASTICS_CONFIG_PATH = "pytest_stochastics.json"
DESIRED_RUNNER_NAME = "pytest_stochastics_runner"


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup("stochastics")
    group.addoption(
        "--plan",
        action="store",
        dest="plan",
        help="Specify the plan name to use for stochastic testing",
    )
    group.addoption(
        "--plan-config-file",
        action="store",
        dest="plan_config_file",
        default=PYTEST_STOCHASTICS_CONFIG_PATH,
        help="Specify a path to the pytest stochastics config file",
    )


def pytest_configure(config: pytest.Config) -> None:
    """Main entry point into the pytest plugin flow."""

    requested_log_level = str(config.getoption("log_cli_level") or "ERROR")
    logger = Logger(name=DESIRED_RUNNER_NAME, level=requested_log_level)
    logger.addHandler(logging.StreamHandler())

    plan_name = str(config.getoption("plan") or "default")
    logger.info(f"Selected Stochastics Plan: {plan_name}")

    config_file_path = str(config.getoption("plan_config_file"))
    logger.info(f"Loading Stochastics Config from: {config_file_path}")

    try:
        runner_config = _load_config(config_file_path)
        logger.debug(f"Loaded runner_config: {runner_config}")
    except Exception as ex:
        logger.error(f"Failed to load runner configuration: {ex}")
        config.add_cleanup(
            lambda: rich.print(
                "\n[red]Stochastic Runner not used, see error log above (look before test session starts) for details.[/red]"
            )
        )
        return

    try:
        runner_selector = RunnerStochastics(PlanId(plan_name), runner_config, logger)
        confirmed_name = config.pluginmanager.register(runner_selector, DESIRED_RUNNER_NAME)
        if confirmed_name is None or confirmed_name != DESIRED_RUNNER_NAME:
            raise Exception(f"Failed to register `{DESIRED_RUNNER_NAME}` plugin!")
        logger.debug(f"Confirmed runner name: {confirmed_name}")
    except Exception as ex:
        logger.error(f"Failed to register runner: {ex}")
        config.add_cleanup(
            lambda: rich.print(
                "\n[red]Stochastic Runner not registered, see error log above (look before test session starts) for details.[/red]"
            )
        )


def _load_config(stochastics_config_path: str) -> RunnerStochasticsConfig:
    with open(stochastics_config_path) as config_file:
        raw_json = config_file.read()
        return RunnerStochasticsConfig.from_json(raw_json)  # type: ignore #TODO?



from dataclasses import dataclass, field
from typing import Any, Dict, NewType, TypeVar

from dataclasses_json import config, dataclass_json  # type: ignore

# Type aliases for keys
PolicyId = NewType("PolicyId", str)
PlanId = NewType("PlanId", str)
TestId = NewType("TestId", str)

T = TypeVar("T", PolicyId, PlanId)


def create_field_metadata(json_field_name: str, field_type: type[T]) -> Dict[str, Any]:
    """Create metadata configuration for field with alternate names"""

    def encoder(val: T) -> str:
        return str(val)

    return config(field_name=json_field_name, encoder=encoder)  # type: ignore


@dataclass_json
@dataclass(frozen=True)
class PolicyTests:
    name: PolicyId = field(metadata=create_field_metadata("policy", PolicyId))
    tests: list[TestId]


@dataclass_json
@dataclass(frozen=True)
class TestPlans:
    name: PlanId = field(metadata=create_field_metadata("plan", PlanId))
    policy_tests: list[PolicyTests]


@dataclass_json
@dataclass(frozen=True)
class Policy:
    name: PolicyId = field(metadata=create_field_metadata("policy", PolicyId))
    at_least: int
    out_of: int


@dataclass_json
@dataclass(frozen=True)
class PlanFallback:
    name: PlanId = field(metadata=create_field_metadata("plan", PlanId))
    overrides: PlanId


@dataclass_json
@dataclass(frozen=True)
class RunnerStochasticsConfig:
    test_plan_list: list[TestPlans] = field(default_factory=list)
    policy_list: list[Policy] = field(default_factory=list)
    plan_fallback_list: list[PlanFallback] = field(default_factory=list)


TestPolicy = dict[TestId, Policy]


def gen_fallback_lookup(
    runner_config: RunnerStochasticsConfig,
    plan: PlanId,
) -> TestPolicy:
    """
    Based on the provided `runner_config` and `plan`,
    generates a lookup from test `nodeid` (`TestId`) to the `Policy` resolved for that test.
    """
    result: TestPolicy = {}

    configured_plans = {bt.name: bt.policy_tests for bt in runner_config.test_plan_list}
    fallback_plans = {fb.name: fb.overrides for fb in runner_config.plan_fallback_list}
    policies = {st.name: st for st in runner_config.policy_list}

    plan_priorities: list[PlanId] = []
    while True:
        if plan in configured_plans:
            plan_priorities.append(plan)
        if plan not in fallback_plans:
            break
        plan = fallback_plans[plan]

    for plan in plan_priorities:
        for policy_tests in configured_plans[plan]:
            for test in policy_tests.tests:
                if test in result:
                    continue
                result[test] = policies[policy_tests.name]

    return result



from logging import Logger
from typing import Generator, cast

# from rich.logging import log
import pytest
from _pytest.reports import TestReport
import rich

from pytest_stochastics.collector import StochasticFunctionCollector
from pytest_stochastics.runner_data import PlanId, RunnerStochasticsConfig, TestId, gen_fallback_lookup


class RunnerStochastics:
    """Stochastics Plugin Runner. Handles the relevant hooks."""

    def __init__(
        self,
        current_plan: str,
        runner_config: RunnerStochasticsConfig,
        logger: Logger,
    ) -> None:
        self.plan = current_plan
        self.runner_config = runner_config
        self._logger = logger

        self.lookup_test_policies = gen_fallback_lookup(runner_config, PlanId(current_plan))

        self._test_result_goals: dict[str, int] = {}
        self._stochastic_items: dict[str, pytest.Item] = {}
        self._flag_any_non_stochastic_test_failed = False

    def pytest_pycollect_makeitem(
        self,
        collector: pytest.Module | pytest.Class,
        name: str,
        obj: object,
    ) -> None | pytest.Item | pytest.Collector | list[pytest.Item | pytest.Collector]:
        """Interecpt `Function`s at collection time and try to match them with a strategy."""

        if not name.startswith("test_"):
            return None

        nodeid = TestId(f"{collector.nodeid}::{name}")

        if nodeid not in self.lookup_test_policies:
            self._logger.debug(f"No policy found for {name}, not wrapping.")
            return None

        test_policy = self.lookup_test_policies[nodeid]

        if test_policy.out_of == 1 and test_policy.at_least > 0:
            self._logger.debug(f"Redundant policy found for {name}, not wrapping")
            return None

        self._logger.info(
            f"Wrapping stochastic function: `{name}` with policy: `{test_policy.name}`[{test_policy.at_least}/{test_policy.out_of}]"
        )

        return StochasticFunctionCollector.from_parent(
            collector,
            name=name,
            obj=obj,
            policy=test_policy,
        )

    @pytest.hookimpl(wrapper=True, trylast=True)
    def pytest_sessionstart(
        self,
        session: pytest.Session,
    ) -> Generator[None, None, None]:
        """Inject around session start to add some common-sense config prints."""

        yield None

        terminal_writer = session.config.get_terminal_writer()
        terminal_writer.write("stochastics plan: ")
        markup: dict[str, bool] = {}

        if len(self.lookup_test_policies) == 0:
            markup["red"] = True
        else:
            markup["green"] = True

        terminal_writer.write(f"{self.plan}", **markup)
        terminal_writer.line(f" [stochastic runs={len(self.lookup_test_policies)}]\n")

    @pytest.hookimpl(wrapper=True)
    def pytest_runtest_protocol(
        self,
        item: pytest.Item,
        nextitem: pytest.Item | None,
    ) -> Generator[None, None, bool | None]:
        """Inject test groupings and summaries around the `pytest_runtest_protocol`."""

        if not isinstance(item.parent, StochasticFunctionCollector):
            yield None
            return None

        key = item.parent.nodeid
        policy = item.parent.policy
        if key not in self._test_result_goals:
            # Starting a new stochastics run
            self._test_result_goals[key] = policy.at_least
            rich.print(
                f"\n\r[blue]StochasticSet:[/blue] {key} [[blue]{policy.name}: {policy.at_least}/{policy.out_of}[/blue]]",
                end="",
            )

        # skip to the end of the protocol
        yield None

        if nextitem is None or item.parent != nextitem.parent:
            # if we're the last item of this stochastics run
            penalty = self._test_result_goals[key]
            comment = f"[yellow]{policy.out_of - (policy.at_least - penalty)} fail(s)[/yellow]"
            if penalty <= 0:
                outcome = "[green]PASSED[/green]"
                comment+= f", [green]{policy.at_least - penalty} pass(es)[/green]"
            else:
                outcome = "[red]FAILED[/red]"
                comment += f", [red]missing {penalty} pass(es)[/red]"
                self._flag_any_non_stochastic_test_failed = True
            rich.print(
                f"\n\r[blue]StochasticSet:[/blue] {outcome} [{comment}]",
                end="",
            )
        return None

    def pytest_runtest_makereport(
        self,
        item: pytest.Item,
        call: pytest.CallInfo[None],
    ) -> None:
        """Called before each step's report is created."""

        if call.when == "call":
            if not isinstance(item.parent, StochasticFunctionCollector):
                return

            # before the actual test function is called, we store it with its parent nodeid.
            self._stochastic_items[item.nodeid] = item

    def pytest_runtest_logreport(
        self,
        report: TestReport,
    ) -> None:
        """Called after each step's test is run and a report has been generated."""

        if report.nodeid not in self._stochastic_items:
            self._flag_any_non_stochastic_test_failed |= not report.passed
            return

        testid = cast(pytest.Item, self._stochastic_items[report.nodeid].parent).nodeid
        if report.when == "call" and report.passed:
            self._test_result_goals[testid] -= 1

    def pytest_sessionfinish(
        self,
        session: pytest.Session,
        exitstatus: int,
    ) -> None:
        if len(self._stochastic_items) == 0:
            return

        if exitstatus != pytest.ExitCode.TESTS_FAILED:
            return

        if any(v > 0 for v in self._test_result_goals.values()) or self._flag_any_non_stochastic_test_failed:
            session.config.add_cleanup(
                lambda: rich.print("\n\r[red]Some tests failed.[/red][blue] Check the details above...[/blue]\n")
            )
            return

        session.exitstatus = pytest.ExitCode.OK  # rectify the exit code
        session.config.add_cleanup(
            lambda: rich.print(
                "\n\r[green]Testing finished successfully.[/green][blue] Some stochastic tests failed[/blue][green] within acceptable margins.[/green]\n"
            )
        )
